@@ -3,6 +3,7 @@ import math
 import operator
 
 import numpy as np
+from dask._task_spec import Task, TaskRef
 from dask.dataframe.dispatch import make_meta, meta_nonempty
 from dask.dataframe.multi import (
     _concat_wrapper,
@@ -11,6 +12,7 @@ from dask.dataframe.multi import (
     merge_chunk,
 )
 from dask.dataframe.shuffle import partitioning_index
+from dask.typing import Key
 from dask.utils import apply, get_default_shuffle_method
 from toolz import merge_sorted, unique
 
@@ -254,13 +256,18 @@ class Merge(Expr):
                 self.broadcast_side == "left"
                 and set(self.right._meta.index.names) == meta_index_names
             ):
-                return self._bcast_right._divisions()
+                if self.right_index:
+                    return self._bcast_right.divisions
+                _npartitions = self._bcast_right.npartitions
             elif (
                 self.broadcast_side == "right"
                 and set(self.left._meta.index.names) == meta_index_names
             ):
-                return self._bcast_left._divisions()
-            _npartitions = max(self.left.npartitions, self.right.npartitions)
+                if self.left_index:
+                    return self._bcast_left.divisions
+                _npartitions = self._bcast_left.npartitions
+            else:
+                _npartitions = max(self.left.npartitions, self.right.npartitions)
 
         else:
             _npartitions = self._npartitions
@@ -282,7 +289,7 @@ class Merge(Expr):
 
         s_method = self.shuffle_method or get_default_shuffle_method()
         if (
-            s_method in ("tasks", "p2p")
+            s_method in ("disk", "tasks", "p2p")
             and self.how in ("inner", "left", "right", "leftsemi")
             and self.how != broadcast_side
             and broadcast is not False
@@ -602,9 +609,14 @@ class HashJoinP2P(Merge, PartitionsFiltered):
         return None
 
     def _layer(self) -> dict:
-        from distributed.shuffle._core import ShuffleId, barrier_key
+        from distributed.shuffle._core import (
+            P2PBarrierTask,
+            ShuffleId,
+            barrier_key,
+            p2p_barrier,
+        )
         from distributed.shuffle._merge import merge_unpack
-        from distributed.shuffle._shuffle import shuffle_barrier
+        from distributed.shuffle._shuffle import DataFrameShuffleSpec
 
         dsk = {}
         token_left = _tokenize_deterministic(
@@ -634,49 +646,80 @@ class HashJoinP2P(Merge, PartitionsFiltered):
         transfer_keys_right = list()
         func = create_assign_index_merge_transfer()
         for i in range(self.left.npartitions):
-            transfer_keys_left.append((transfer_name_left, i))
-            dsk[(transfer_name_left, i)] = (
+            t = Task(
+                (transfer_name_left, i),
                 func,
-                (self.left._name, i),
+                TaskRef((self.left._name, i)),
                 self.shuffle_left_on,
                 _HASH_COLUMN_NAME,
                 self.npartitions,
                 token_left,
                 i,
-                self.left._meta,
-                self._partitions,
                 self.left_index,
             )
+            dsk[t.key] = t
+            transfer_keys_left.append(t.ref())
+
         for i in range(self.right.npartitions):
-            transfer_keys_right.append((transfer_name_right, i))
-            dsk[(transfer_name_right, i)] = (
+            t = Task(
+                (transfer_name_right, i),
                 func,
-                (self.right._name, i),
+                TaskRef((self.right._name, i)),
                 self.shuffle_right_on,
                 _HASH_COLUMN_NAME,
                 self.npartitions,
                 token_right,
                 i,
-                self.right._meta,
-                self._partitions,
                 self.right_index,
             )
+            dsk[t.key] = t
+            transfer_keys_right.append(t.ref())
 
-        dsk[_barrier_key_left] = (shuffle_barrier, token_left, transfer_keys_left)
-        dsk[_barrier_key_right] = (
-            shuffle_barrier,
+        meta_left = self.left._meta.assign(**{_HASH_COLUMN_NAME: 0})
+        barrier_left = P2PBarrierTask(
+            _barrier_key_left,
+            p2p_barrier,
+            token_left,
+            transfer_keys_left,
+            spec=DataFrameShuffleSpec(
+                id=token_left,
+                npartitions=self.npartitions,
+                column=_HASH_COLUMN_NAME,
+                meta=meta_left,
+                parts_out=self._partitions,
+                disk=True,
+                drop_column=True,
+            ),
+        )
+        dsk[barrier_left.key] = barrier_left
+
+        meta_right = self.right._meta.assign(**{_HASH_COLUMN_NAME: 0})
+        barrier_right = P2PBarrierTask(
+            _barrier_key_right,
+            p2p_barrier,
             token_right,
             transfer_keys_right,
+            spec=DataFrameShuffleSpec(
+                id=token_right,
+                npartitions=self.npartitions,
+                column=_HASH_COLUMN_NAME,
+                meta=meta_right,
+                parts_out=self._partitions,
+                disk=True,
+                drop_column=True,
+            ),
         )
+        dsk[barrier_right.key] = barrier_right
 
         for part_out in self._partitions:
-            dsk[(self._name, part_out)] = (
+            t = Task(
+                (self._name, part_out),
                 merge_unpack,
                 token_left,
                 token_right,
                 part_out,
-                _barrier_key_left,
-                _barrier_key_right,
+                barrier_left.ref(),
+                barrier_right.ref(),
                 self.how,
                 self.left_on,
                 self.right_on,
@@ -686,6 +729,7 @@ class HashJoinP2P(Merge, PartitionsFiltered):
                 self.right_index,
                 self.indicator,
             )
+            dsk[t.key] = t
         return dsk
 
     def _simplify_up(self, parent, dependents):
@@ -718,8 +762,14 @@ class BroadcastJoin(Merge, PartitionsFiltered):
 
     def _divisions(self):
         if self.broadcast_side == "left":
-            return self.right._divisions()
-        return self.left._divisions()
+            if self.right_index:
+                return self.right.divisions
+            npartitions = self.right.npartitions
+        else:
+            if self.left_index:
+                return self.left.divisions
+            npartitions = self.left.npartitions
+        return (None,) * (npartitions + 1)
 
     def _simplify_up(self, parent, dependents):
         return
@@ -792,7 +842,6 @@ class BroadcastJoin(Merge, PartitionsFiltered):
 
 
 def create_assign_index_merge_transfer():
-    import pandas as pd
     from distributed.shuffle._core import ShuffleId
     from distributed.shuffle._merge import merge_transfer
 
@@ -803,8 +852,6 @@ def create_assign_index_merge_transfer():
         npartitions,
         id: ShuffleId,
         input_partition: int,
-        meta: pd.DataFrame,
-        parts_out: set[int],
         index_merge,
     ):
         if index_merge:
@@ -826,10 +873,7 @@ def create_assign_index_merge_transfer():
 
         index = partitioning_index(index, npartitions)
         df = df.assign(**{name: index})
-        meta = meta.assign(**{name: 0})
-        return merge_transfer(
-            df, id, input_partition, npartitions, meta, parts_out, True
-        )
+        return merge_transfer(df, id, input_partition)
 
     return assign_index_merge_transfer
 
@@ -880,17 +924,15 @@ class BlockwiseMerge(Merge, Blockwise):
     def _broadcast_dep(self, dep: Expr):
         return dep.npartitions == 1
 
-    def _task(self, index: int):
+    def _task(self, name: Key, index: int) -> Task:
         kwargs = self.kwargs.copy()
         kwargs["result_meta"] = self._meta
-        return (
-            apply,
+        return Task(
+            name,
             merge_chunk,
-            [
-                self._blockwise_arg(self.left, index),
-                self._blockwise_arg(self.right, index),
-            ],
-            kwargs,
+            self._blockwise_arg(self.left, index),
+            self._blockwise_arg(self.right, index),
+            **kwargs,
         )
 
 
